# -*- coding: utf-8 -*-


class RemoteClient(object):

    def __init__(self):
        pass

    def connect(self, endpoint):
        pass

    def send(self, message):
        pass

    def recv(self, timeout=None):
        pass

